﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.CombinedIntegerVectorEncoding;
using HeuristicLab.Encodings.ConditionActionEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.Instances;

namespace HeuristicLab.Problems.ConditionActionClassification {
  [StorableClass]
  public class ConditionActionClassificationProblem : HeuristicOptimizationProblem<XCSEvaluator, UniformRandomCombinedIntegerVectorCreator>, IConditionActionProblem,
  IProblemInstanceConsumer<ConditionActionClassificationProblemData> {
    private const string ClassifierFetcherParameterName = "ClassifierFetcher";
    private const string ActionExecuterParameterName = "ActionExecuter";
    private const string ActionSetSubsumptionOperatorParameterName = "ActionSetSubsumption";

    IXCSEvaluator IConditionActionProblem.Evaluator {
      get { return Evaluator; }
    }

    #region parameter properties
    public IValueParameter<ConditionActionClassificationProblemData> ProblemDataParameter {
      get { return (IValueParameter<ConditionActionClassificationProblemData>)Parameters["ProblemData"]; }
    }
    public IValueParameter<ICoveringSolutionCreator> CoveringSolutionCreatorParameter {
      get { return (IValueParameter<ICoveringSolutionCreator>)Parameters["CoveringSolutionCreator"]; }
    }
    public IFixedValueParameter<PercentValue> ChangeSymbolProbabilityInCoveringParameter {
      get { return (IFixedValueParameter<PercentValue>)Parameters["ChangeSymbolProbabilityInCovering"]; }
    }
    public IFixedValueParameter<DoubleValue> PositiveRewardParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["PositiveReward"]; }
    }
    public IFixedValueParameter<DoubleValue> NegativeRewardParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["NegativeReward"]; }
    }
    public IFixedValueParameter<DoubleValue> InitialPredictionParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["InitialPrediction"]; }
    }
    public IFixedValueParameter<DoubleValue> InitialErrorParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["InitialError"]; }
    }
    public IFixedValueParameter<DoubleValue> InitialFitnessParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["InitialFitness"]; }
    }
    public IFixedValueParameter<ItemSet<IAction>> PossibleActionsParameter {
      get { return (IFixedValueParameter<ItemSet<IAction>>)Parameters["PossibleActions"]; }
    }
    public IFixedValueParameter<ItemSet<CombinedIntegerVector>> PossibleActionsConcreteClassParameter {
      get { return (IFixedValueParameter<ItemSet<CombinedIntegerVector>>)Parameters["PossibleActionsConcreteClass"]; }
    }
    public IFixedValueParameter<IntValue> ThetaMinimalNumberOfActionsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters["ThetaMinimalNumberOfActions"]; }
    }
    public IFixedValueParameter<HeuristicLab.Encodings.CombinedIntegerVectorEncoding.CombinedIntegerVector.CombinedIntegerVectorComparer> ClassifierComparerParameter {
      get { return (IFixedValueParameter<HeuristicLab.Encodings.CombinedIntegerVectorEncoding.CombinedIntegerVector.CombinedIntegerVectorComparer>)Parameters["ClassifierComparer"]; }
    }
    #endregion

    #region properties
    IParameter IConditionActionProblem.ProblemDataParameter {
      get { return ProblemDataParameter; }
    }
    IConditionActionProblemData IConditionActionProblem.ProblemData {
      get { return ProblemData; }
    }
    public ConditionActionClassificationProblemData ProblemData {
      get { return ProblemDataParameter.Value; }
      protected set {
        ProblemDataParameter.Value = value;
        if (value != null) {
          SetProblemDataSpecificParameters();
        }
      }
    }
    IParameter IConditionActionProblem.PossibleActionsConcreteClassParameter {
      get { return PossibleActionsConcreteClassParameter; }
    }
    public ItemSet<CombinedIntegerVector> PossibleActionsConcreteClass {
      get { return PossibleActionsConcreteClassParameter.Value; }
    }
    IParameter IConditionActionProblem.PossibleActionsParameter {
      get { return PossibleActionsParameter; }
    }
    //IItemSet<IClassifier> IConditionActionProblem.PossibleActions {
    //  get { return PossibleActions; }
    //}
    public ItemSet<IAction> PossibleActions {
      get { return PossibleActionsParameter.Value; }
    }
    public IActionExecuter ActionExecuter {
      get { return ActionExecuterParameter.Value; }
    }
    public ValueParameter<IActionExecuter> ActionExecuterParameter {
      get { return (ValueParameter<IActionExecuter>)Parameters[ActionExecuterParameterName]; }
    }
    IParameter IConditionActionProblem.ActionExecuterParameter { get { return ActionExecuterParameter; } }
    public ClassifierFetcher ClassifierFetcher {
      get { return ClassifierFetcherParameter.Value; }
    }
    public ValueParameter<ClassifierFetcher> ClassifierFetcherParameter {
      get { return (ValueParameter<ClassifierFetcher>)Parameters[ClassifierFetcherParameterName]; }
    }
    IClassifierFetcher IConditionActionProblem.ClassifierFetcher { get { return ClassifierFetcher; } }
    IParameter IConditionActionProblem.ClassifierFetcherParameter { get { return ClassifierFetcherParameter; } }


    public ActionSetSubsumptionOperator ActionSetSubsumptionOperator {
      get { return ActionSetSubsumptionOperatorParameter.Value; }
    }
    public ValueParameter<ActionSetSubsumptionOperator> ActionSetSubsumptionOperatorParameter {
      get { return (ValueParameter<ActionSetSubsumptionOperator>)Parameters[ActionSetSubsumptionOperatorParameterName]; }
    }
    IActionSetSubsumption IConditionActionProblem.ActionSetSubsumptionOperator { get { return ActionSetSubsumptionOperator; } }
    IParameter IConditionActionProblem.ActionSetSubsumptionOperatorParameter { get { return ActionSetSubsumptionOperatorParameter; } }

    private IntValue ThetaMinimalNumberOfActions {
      get { return ThetaMinimalNumberOfActionsParameter.Value; }
    }
    IParameter IConditionActionProblem.ThetaMinimalNumberOfActionsParameter {
      get { return ThetaMinimalNumberOfActionsParameter; }
    }
    public ICoveringSolutionCreator CoveringSolutionCreator {
      get { return CoveringSolutionCreatorParameter.Value; }
    }
    IParameter IConditionActionProblem.CoveringSolutionCreatorParameter {
      get { return CoveringSolutionCreatorParameter; }
    }
    public IClassifierComparer ClassifierComparer {
      get { return ClassifierComparerParameter.Value; }
    }
    IParameter IConditionActionProblem.ClassifierComparerParameter {
      get { return ClassifierComparerParameter; }
    }
    #endregion

    [StorableConstructor]
    protected ConditionActionClassificationProblem(bool deserializing) : base(deserializing) { }
    protected ConditionActionClassificationProblem(ConditionActionClassificationProblem original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ConditionActionClassificationProblem(this, cloner);
    }

    public ConditionActionClassificationProblem() :
      this(new ConditionActionClassificationProblemData(new Dataset(ConditionActionClassificationProblemData.defaultVariableNames, ConditionActionClassificationProblemData.defaultData),
        ConditionActionClassificationProblemData.defaultVariableNames.Take(ConditionActionClassificationProblemData.defaultVariableNames.Length - 1), ConditionActionClassificationProblemData.defaultVariableNames.Last().ToEnumerable()),
        new XCSEvaluator(), new UniformRandomCombinedIntegerVectorCreator(), new CombinedIntegerVectorCoveringCreator()) {
    }

    public ConditionActionClassificationProblem(ConditionActionClassificationProblemData problemData, XCSEvaluator evaluator, UniformRandomCombinedIntegerVectorCreator solutionCreator, ICoveringSolutionCreator coveringSolutionCreator)
      : base(evaluator, solutionCreator) {
      Parameters.Add(new ValueParameter<ConditionActionClassificationProblemData>("ProblemData", "", problemData));
      Parameters.Add(new FixedValueParameter<DoubleValue>("PositiveReward", "", new DoubleValue(1000)));
      Parameters.Add(new FixedValueParameter<DoubleValue>("NegativeReward", "", new DoubleValue(0)));
      Parameters.Add(new FixedValueParameter<DoubleValue>("InitialPrediction", "Initial Presiction", new DoubleValue(0)));
      Parameters.Add(new FixedValueParameter<DoubleValue>("InitialError", "Initial Error", new DoubleValue(0)));
      Parameters.Add(new FixedValueParameter<DoubleValue>("InitialFitness", "Initial Fitness", new DoubleValue(0)));

      Parameters.Add(new ValueParameter<IActionExecuter>(ActionExecuterParameterName, "", new ActionExecuter()));
      Parameters.Add(new ValueParameter<ClassifierFetcher>(ClassifierFetcherParameterName, "", new ClassifierFetcher()));
      Parameters.Add(new FixedValueParameter<HeuristicLab.Encodings.CombinedIntegerVectorEncoding.CombinedIntegerVector.CombinedIntegerVectorComparer>("ClassifierComparer", new HeuristicLab.Encodings.CombinedIntegerVectorEncoding.CombinedIntegerVector.CombinedIntegerVectorComparer()));
      Parameters.Add(new FixedValueParameter<ItemSet<IAction>>("PossibleActions", new ItemSet<IAction>(ClassifierComparer)));
      Parameters.Add(new FixedValueParameter<ItemSet<CombinedIntegerVector>>("PossibleActionsConcreteClass", new ItemSet<CombinedIntegerVector>(ClassifierComparer)));
      Parameters.Add(new FixedValueParameter<IntValue>("ThetaMinimalNumberOfActions", "Minimal number of actions, which have to be present in the match set, or else covering will occure.", new IntValue(1)));

      Parameters.Add(new ValueParameter<ICoveringSolutionCreator>("CoveringSolutionCreator", "", coveringSolutionCreator));
      Parameters.Add(new FixedValueParameter<PercentValue>("ChangeSymbolProbabilityInCovering", "", new PercentValue(0.33)));

      Parameters.Add(new ValueParameter<ActionSetSubsumptionOperator>(ActionSetSubsumptionOperatorParameterName, "", new ActionSetSubsumptionOperator()));

      Evaluator.InitialErrorParameter.ActualName = "InitialError";
      Evaluator.InitialFitnessParameter.ActualName = "InitialFitness";
      Evaluator.InitialPredictionParameter.ActualName = "InitialPrediction";

      coveringSolutionCreator.ChangeSymbolProbabilityParameter.ActualName = ChangeSymbolProbabilityInCoveringParameter.Name;
      coveringSolutionCreator.CoverInputParameter.ActualName = ClassifierFetcher.CurrentInputToMatchParameter.ActualName;
      coveringSolutionCreator.CreatedClassifierParameter.ActualName = "CombinedIntegerVector";

      ClassifierFetcher.ProblemDataParameter.ActualName = ProblemDataParameter.Name;

      ActionExecuter.CurrentActionToMatchParameter.ActualName = ClassifierFetcher.CurrentActionToMatchParameter.ActualName;
      ActionExecuter.NegativeRewardParameter.ActualName = NegativeRewardParameter.Name;
      ActionExecuter.PositiveRewardParameter.ActualName = PositiveRewardParameter.Name;

      ActionSetSubsumptionOperator.ClassifiersParameter.ActualName = "CombinedIntegerVector";

      SetProblemDataSpecificParameters();
      ThetaMinimalNumberOfActions.Value = PossibleActions.Count;

      InitializeOperators();

      problemData.Changed += new System.EventHandler(problemData_Changed);
    }

    private void problemData_Changed(object sender, System.EventArgs e) {
      SetProblemDataSpecificParameters();
    }

    private void SetProblemDataSpecificParameters() {
      SolutionCreator.ActionPartLengthParameter.ActualName = ProblemData.ActionLengthParameter.Name;
      SolutionCreator.LengthParameter.ActualName = ProblemData.LengthParameter.Name;
      SolutionCreator.BoundsParameter.ActualName = ProblemData.BoundsParameter.Name;

      SetPossibleActions();
    }

    private void InitializeOperators() {
      Operators.AddRange(ApplicationManager.Manager.GetInstances<ICombinedIntegerVectorCrossover>());
      Operators.AddRange(AddManipulators());
      Operators.Add(new BestTrainingXCSSolutionAnalyzer());
      Operators.Add(new CurrentXCSSolutionAnalyzer());

      ParameterizeOperators();
    }

    private IEnumerable<ICombinedIntegerVectorManipulator> AddManipulators() {
      var manipulator = new UniformSomePositionManipulator();
      manipulator.ChildParameter.ActualName = "CombinedIntegerVector";
      manipulator.FetchedInputParameter.ActualName = ClassifierFetcher.CurrentInputToMatchParameter.ActualName;
      manipulator.PossibleActionsParameter.ActualName = PossibleActionsConcreteClassParameter.Name;
      return new List<ICombinedIntegerVectorManipulator>() { manipulator };
    }

    private void ParameterizeOperators() {
      foreach (XCSSolutionAnalyzer xcsAnalyzer in Operators.OfType<XCSSolutionAnalyzer>()) {
        xcsAnalyzer.ClassifierParameter.ActualName = SolutionCreator.CombinedIntegerVectorParameter.ActualName;
        xcsAnalyzer.PredictionParameter.ActualName = Evaluator.PredictionParameter.ActualName;
        xcsAnalyzer.ErrorParameter.ActualName = Evaluator.ErrorParameter.ActualName;
        xcsAnalyzer.FitnessParameter.ActualName = Evaluator.FitnessParameter.ActualName;
        xcsAnalyzer.ExperienceParameter.ActualName = Evaluator.ExperienceParameter.ActualName;
        xcsAnalyzer.AverageActionSetSizeParameter.ActualName = Evaluator.AverageActionSetSizeParameter.ActualName;
        xcsAnalyzer.NumerosityParameter.ActualName = Evaluator.NumerosityParameter.ActualName;
        xcsAnalyzer.TimestampParameter.ActualName = Evaluator.TimestampParameter.ActualName;
        xcsAnalyzer.ProblemDataParameter.ActualName = ProblemDataParameter.Name;
        xcsAnalyzer.ResultsParameter.ActualName = "Results";
        xcsAnalyzer.ClassifierComparerParameter.ActualName = ClassifierComparerParameter.Name;
      }
    }

    private void SetPossibleActions() {
      //get bounds of action
      IntMatrix actionBounds = GetElementsOfBoundsForAction(ProblemData.Bounds, ProblemData.Length.Value, ProblemData.ActionLength.Value);
      int actionLength = ProblemData.ActionLength.Value;
      int start = ProblemData.Length.Value - actionLength;
      int[] elements = new int[actionLength];
      int[] curPos = new int[actionLength];
      bool done = false;
      //initialize curPos
      for (int i = 0; i < actionBounds.Rows; i++) {
        curPos[i] = actionBounds[i, 0];
      }
      PossibleActions.Clear();
      PossibleActionsConcreteClass.Clear();
      while (!done) {
        PossibleActions.Add(new CombinedIntegerVector(curPos, actionLength, actionBounds));
        PossibleActionsConcreteClass.Add(new CombinedIntegerVector(curPos, actionLength, actionBounds));
        curPos = GetNextAction(curPos, actionBounds, out done);
      }

    }

    private int[] GetNextAction(int[] curPos, IntMatrix actionBounds, out bool done) {
      int cur = 0;
      while (cur < curPos.Length) {
        curPos[cur] += actionBounds.Columns < 3 ? 1 : actionBounds[cur, 2];
        if (curPos[cur] >= actionBounds[cur, 1]) {
          curPos[cur] = actionBounds[cur, 0];
          cur++;
        } else {
          break;
        }
      }
      done = cur >= curPos.Length;
      return curPos;
    }

    private IntMatrix GetElementsOfBoundsForAction(IntMatrix bounds, int length, int actionPartLength) {
      IntMatrix actionBounds = new IntMatrix(actionPartLength, bounds.Columns);
      int start = length - actionPartLength;
      for (int i = start; i < length; i++) {
        int pos = i % bounds.Rows;
        for (int j = 0; j < bounds.Columns; j++) {
          actionBounds[i - start, j] = bounds[pos, j];
        }
      }
      return actionBounds;
    }

    public void Load(ConditionActionClassificationProblemData data) {
      Name = data.Name;
      Description = data.Description;
      ProblemData = data;
    }
  }
}
